
-- local thc= math.sqrt(self.radius^2 - dq) removed to save gpu
-- local t1 = tc - thc or tc + thc

raytracer_shader = [[
 #define csphere 4
 #define error   1000000.0f

 //window size stuff
 float width  = 800;
 float height = 640;
 float aspect = width/height;
 float fov    = tan(3.141 * 0.5 * 65 / 180);//fov of 65

 struct tsphere{
  vec4  specular;
  vec4  color;
  vec3  position; 
  float radius;
  float reflection;
  float fresnelstr;
 };

 struct tlight{
  vec3  direction;
  vec3  position;
  float strenght;
 };

 struct tcamera{
  vec3 position;
  vec3 lookat;
 };

 uniform tsphere sphere[csphere];// our geometry
 uniform tlight  light;
 uniform tcamera camera;
 
 float planeintersect(vec3 pos,vec3 dir,vec4 plane){
  if(dot(vec3(plane),dir) < 0){return -1.0;}//check if they are perpendicular

  vec3 normal = vec3(plane);
  return (dot(pos,normal)+plane.w) / (dot(dir,normal));
 }


 float sphereintersect(vec3 pos,vec3 dir,vec3 origin,float rad,out float t2){
  vec3   L = origin - pos;
  float tc = dot(L,dir);
  if(tc < 0.0){return -1.0;}
  float dq = dot(L,L) - tc*tc;
  if(dq > rad*rad){return -1.0;}
  float thc= sqrt(rad*rad - dq);
  float t1 = tc - thc;
  t2 = tc + thc;

  return t1;
 }

 float getClosestIntersect(vec3 pos,vec3 dir,out int index,out float t2){
  float closest = error;
  float closest1= 0;
  t2      = -1;

  for(int b = 0;b<csphere;b++){
   float inters = sphereintersect(pos,dir,sphere[b].position ,sphere[b].radius,t2);
   if(inters > -1.0 && closest > inters){
    closest = inters;
    closest1= t2;
    index   = b;
   }
  }
  t2 = closest1;

  vec3 plane_normal = vec3(0,1,0);
  float inters = planeintersect(pos,dir,vec4(plane_normal,0.5 - camera.position.y*2 ) );
  if(inters > -1.0 && closest > inters){
   closest = inters;
   index   = -1;
  }

  return closest;
 }

/* decided to leave the current reflect function
 vec3 getNormal(vec3 hit,int index){
  if(index == -1){return vec3(0,1,0);}
  return normalize(hit - sphere[index].position);
 }

 vec4 getColor(int index){
  if(index == -1){return vec4(0.5);}
  return sphere[index].color;
 } 


 vec4 reflectionIntersect(int index,int loop,vec3 pos,vec3 dir,out vec4 reflcolor){
  float temp = 0;
  if(loop == 0){return vec4(1.0);}

  int   ind = -1;
  int   ind2= -1;
  vec4  color= vec4(1.0);

  float hit = getClosestIntersect(pos,dir,ind,temp);
  if(hit != error){
   float theta = clamp(dot(normalize(light.position - (pos + dir * hit)),getNormal(pos + dir * hit,ind)),0,1);
   vec4  hcol  = getColor(ind);
   float shad  = getClosestIntersect((pos + dir * hit),normalize(light.position - (pos + dir * hit)),ind2,temp);//from hit to light
   if(shad != error){
    shad       = 0.3f;
   }else{
    shad       = 1.0f;
   } 

   if(ind != -1 && sphere[ind].reflection > 0.1){
    reflcolor = hcol * shad * theta;
    //reflectionIntersect(loop - 1)
    return reflcolor;
   }
   

  }

  return color;
 }
*/

 vec4 effect(vec4 c, Image tex, vec2 tc, vec2 sc){
  float rayx   = (2     * ((sc.x + 0.5) / width ) - 1) * fov * aspect;
  float rayy   = (1 - 2 * ((sc.y + 0.5) / height))     * fov;

  vec3  dirz   = normalize(camera.lookat - camera.position);
  vec3  dirx   = cross(vec3(0,1,0),dirz);
  vec3  diry   = cross(dirz,dirx);

  mat3  view   = mat3(dirx.x,diry.x,dirz.x,dirx.y,diry.y,dirz.y,dirx.z,diry.z,dirz.z);
  vec3  dir    = normalize(vec3(rayx,rayy, 1)  ) * view;

  vec4  color  = vec4(1.0);
  float shadow = 1.0f;
  float distanc= 1.0f;
  float theta  = 1.0f;
  int   index  = -1;
  vec4  fresnel= vec4(1.0);

  float temp = 0;
  float dist2= 0;
  float dist = getClosestIntersect(camera.position,dir,index,dist2);

  if(dist != error){//we hit something
   if(index != -1){color = sphere[index].color;}
   else{color = vec4(0.5);}//plane should be gray

   vec3 hitpoint = camera.position + dir*dist;
   int new_index = -1;
   distanc       = distance(light.position,hitpoint);

   dist = getClosestIntersect(hitpoint,normalize(light.position - hitpoint),new_index,temp); //shoot a shadow ray means from hitpoint to light 
   if(dist != error && new_index != index){//we hit something so we are in shadow new_index != index to avoid self shading
    shadow          = 0.3f;
    //if(new_index != -1){shadow = shadow + sphere[new_index].reflection/4;}
   }else{
    shadow= 1.0f;
   }

   vec3 lightnormal= normalize(light.position - hitpoint);
   vec3 normal     = normalize(hitpoint - sphere[index].position);if(index == -1){normal = vec3(0,1,0);} 

   if(index != -1 && sphere[index].reflection < 0.1){  
    theta = clamp(dot(lightnormal,normal),0,1);
   }else if(index != -1){//let's reflect 
    
    int new_dist2 = -1;
    float refldist = getClosestIntersect(hitpoint,normal,new_dist2,temp);//from normal to anything
    if(refldist != error){//handle object reflection
     int new_dist3 = -1;
     vec3 reflecthit   = hitpoint + normal * refldist;
     float reflshadist = getClosestIntersect(reflecthit,normalize(light.position - reflecthit),new_dist3,temp);
     if(reflshadist != error){shadow = 0.3f;}else{shadow = 0.6f;}

     vec4 reflcolor = vec4(1.0);if(new_dist2 != -1){reflcolor = sphere[new_dist2].color;}
     vec4  rcolor   = vec4(1.0);
     vec3  inside   = vec3(0,-1,0);if(new_dist2 != -1){inside = normalize(reflecthit - sphere[new_dist2].position);rcolor = sphere[new_dist2].specular;}
     float theta2   = dot(inside,normalize(light.position - reflecthit));

     color = reflcolor*shadow*theta2 + rcolor*shadow*theta2;
    }//reflect backside
    vec3 backside = camera.position + dir * dist2;
    normal        = normalize(backside - sphere[index].position);
    float backdist= getClosestIntersect(backside,normal,new_dist2,temp);
    if(backdist != error){
     int new_dist3 = -1;
     vec3 reflecthit2   = backside + normal * backdist;
     float reflshadist2 = getClosestIntersect(reflecthit2,normalize(light.position - reflecthit2),new_dist3,temp);
     if(reflshadist2 != error){shadow = 0.3f;}else{shadow = 0.6f;}

     vec4 reflcolor = vec4(1.0);if(new_dist2 != -1){reflcolor =  sphere[new_dist2].color;}
     vec4  rcolor   = vec4(1.0);
     vec3  inside   = vec3(0,-1,0);if(new_dist2 != -1){inside = normalize(reflecthit2 - sphere[new_dist2].position);rcolor = sphere[new_dist2].specular;}
     float theta2   = dot(inside,normalize(light.position - reflecthit2));

     color = reflcolor*shadow*theta2 + rcolor*shadow*theta2;
    }
 
    //lastly the fresnel
    normal        = normalize(hitpoint - sphere[index].position);  
    float fresang = -dot(dir,normal) + 0.1;
    fresnel       = mix(vec4(1.0),color,fresang * sphere[index].fresnelstr);
    color         = mix(color,fresnel,0.5);
   }


  }

  vec4 spec = vec4(0.08);if(index != -1){sphere[index].specular;}
  return 0.4 + //ambient
  ((color * theta * light.strenght / (distanc*distanc)) + //diffuse
  (spec   * theta * light.strenght / (distanc*distanc))) * shadow; //specular; 
 }

 
]]
